/*! \file
Copyright (c) 2003, The Regents of the University of California, through
Lawrence Berkeley National Laboratory (subject to receipt of any required 
approvals from U.S. Dept. of Energy) 

All rights reserved. 

The source code is distributed under BSD license, see the file License.txt
at the top-level directory.
*/
/*! \mainpage SuperLU_DIST Documentation
 
  SuperLU_DIST is a parallel library for the direct solution of large,
  sparse, nonsymmetric systems of linear equations for distributed
  memory machines. The library is written in C and MPI, and is callable
  from either C or Fortran. The library routines perform an LU
  decomposition with static pivoting and triangular system solutions
  through forward and back substitution. 
 
 */
